#include "structs.h"
#include <inttypes.h> 
#include <avr/interrupt.h>



typedef void ( *ReceiveFunctionPointer ) ( uint8_t );
typedef void ( *SendFunctionPointer ) ( void );
volatile ReceiveFunctionPointer RxDJumpAddress;
volatile SendFunctionPointer TxDJumpAddress;






volatile uint8_t RS232TimeOutWatchdog = 15; 
volatile uint8_t RxDActualTelegrammID;
volatile uint8_t RxDActualCommand;
volatile uint8_t* RxDMemTargetAddress;
volatile uint8_t RxDPeripherieID;
volatile uint16_t  RxDDataLength;

volatile uint8_t RxDChecksum;
volatile uint8_t RxDReadWriteDeviceStatus; 	//Used to signal success of ReadDevice/ WriteDevice commands

#define MSK_RxD_RWDSError 0x40		// 0: SUCCEEDED, 0x40: ERROR
#define  PT_LOCK_TWICurrentWrite_Mask  0x01	// Bit 0: TWICurrentWrite (Locked state for RS232)
#define  PT_LOCK_RS232CurrentWrite_Mask 0x02 	// Bit 1: RS232CurrentWrite (Locked state for TWI)
#define  PT_LOCK_NewWriteData_Mask 0x04		// Bit 2: NewWriteData
#define  PT_LOCK_TWICurrentRead_Mask 0x08		// Bit 3: TWICurrentRead (Locked state for RS232)
#define  PT_LOCK_RS232CurrentRead_Mask 0x10	// Bit 4: RS232CurrentRead (Locked state for RS232)



//Forward declarations
void RxDGetTelegrammID(uint8_t receivedByte);
void RxDInterruptService(uint8_t receivedByte);
void TxDInterruptService();
void RefreshRS232TimeOutWatchdog();
void TestRS232TimeOutWatchdog();
void RxDGetTelegrammID(uint8_t receivedByte);
void RxDGetCommand(uint8_t receivedByte);
void TxDSendCommandAck();
void TxDSendCommandChecksum();
void TxDSendDummy();
PeripherieTableEntry* GetPeripherieTableEntry(uint8_t ID);
void RxDCommandReadMemGetFirstTargetByte(uint8_t receivedByte);
void RxDCommandReadMemGetSecondTargetByte(uint8_t receivedByte);
void RxDCommandReadMemGetFirstLengthByte(uint8_t receivedByte);
void RxDCommandReadMemGetSecondLengthByte(uint8_t receivedByte);
void TxDSendCommandReadMemData();
void RxDCommandWriteMemGetFirstTargetByte(uint8_t receivedByte);
void RxDCommandWriteMemGetSecondTargetByte(uint8_t receivedByte);
void RxDCommandWriteMemGetFirstLengthByte(uint8_t receivedByte);
void RxDCommandWriteMemGetSecondLengthByte(uint8_t receivedByte);
void RxDCommandWriteMemGetData(uint8_t receivedByte);
void RxDCommandReadDeviceGetPeripherieID(uint8_t receivedByte);
void RxDCommandReadDeviceGetFirstLengthByte(uint8_t receivedByte);
void RxDCommandReadDeviceGetSecondLengthByte(uint8_t receivedByte);
void TxDSendCommandReadDeviceData();
void RxDCommandWriteDeviceGetPeripherieID(uint8_t receivedByte);
void RxDCommandWriteDeviceGetFirstLengthByte(uint8_t receivedByte);
void RxDCommandWriteDeviceGetSecondLengthByte(uint8_t receivedByte);
void RxDCommandWriteDeviceGetData(uint8_t receivedByte);
void TxDSendByte(uint8_t sendByte);








//***********************************************************************	
// INIT
//***********************************************************************
void UARTTelegramInit()
{
	RxDJumpAddress = RxDGetTelegrammID;
	TxDJumpAddress = TxDSendDummy;
}

//***********************************************************************	
// Called from RS232.asm (RxD Interrupt vector)
//***********************************************************************	
void RxDInterruptService(uint8_t receivedByte)
{
	RefreshRS232TimeOutWatchdog();
	RxDJumpAddress(receivedByte);
}

//***********************************************************************	
// Called from RS232.asm (TxD Interrupt vector)
//***********************************************************************	
void TxDInterruptService()
{
	RefreshRS232TimeOutWatchdog();
	TxDJumpAddress();
}		



//***********************************************************************	
// RefreshRS232TimeOutWatchdog: 
// Called from subroutines in this module.
// Signals a running data transfer by initializing a watchdog counter
// if watchdog counter (decremented by MainTimer.asm) is timed out, 
//	RxD/TxD state machines are switcheing to 1st State
//***********************************************************************	
void RefreshRS232TimeOutWatchdog()
{
	RS232TimeOutWatchdog = 15;
} 



//***********************************************************************	
// TestRS232TimeOutWatchdog: 
// Called from MainTimerInterrupt: If Counter is timed out, 
//	RxD/TxD state machines are switcheing to 1st State 
//***********************************************************************	
void TestRS232TimeOutWatchdog()
{
	if (RS232TimeOutWatchdog > 0)
	{
		RS232TimeOutWatchdog--;
		if (0 == RS232TimeOutWatchdog)
			UARTTelegramInit();
	}
}

//********************************************************************
// RxD 1st Status: a TelegramID(TicketID to return in answer) is expected
//********************************************************************
void RxDGetTelegrammID(uint8_t receivedByte)
{
	RxDActualTelegrammID = receivedByte;
	RxDJumpAddress = RxDGetCommand;
}

//********************************************************************
// RxD 2nd Status: a Command is expected:
//	1 = WriteMem
//	2 = ReadMem
//  3 = WriteDevice	
//	4 = ReadDevice
//********************************************************************
void RxDGetCommand(uint8_t receivedByte)
{
	RxDActualCommand =  receivedByte; // Save actual command
	
	//Switch command table
	switch(receivedByte)
	{
		case 1:
			RxDJumpAddress = RxDCommandWriteMemGetFirstTargetByte;
			break;
		case 2:
			RxDJumpAddress = RxDCommandReadMemGetFirstTargetByte;
			break;
		case 3:
			RxDJumpAddress = RxDCommandWriteDeviceGetPeripherieID;
			break;
		case 4:
			RxDJumpAddress = RxDCommandReadDeviceGetPeripherieID;
			break;
		default:
			RxDJumpAddress = RxDGetTelegrammID;
	}
}
//********************************************************************
// TxD 2nd Status: decision what kind of answer is returned.
// Remarks: 1st Status (answering with telegramID) is done prior 
//	by RxD subroutine)
//********************************************************************
void TxDSendCommandAck()
{
	switch(RxDActualCommand)
	{
		case 0x01: //WriteMem
			TxDJumpAddress = TxDSendCommandChecksum;
			TxDSendByte(0x81);
			break;
		case 0x02: //ReadMem
			TxDJumpAddress = TxDSendCommandReadMemData;
			TxDSendByte(0x82);
			break;
		case 0x03:	//WriteDevice
			TxDJumpAddress = TxDSendCommandChecksum;
			TxDSendByte(0x83 | RxDReadWriteDeviceStatus);
			break;
		case 0x04: //ReadDevice
			TxDJumpAddress = TxDSendCommandReadDeviceData;
			TxDSendByte(0x84 | RxDReadWriteDeviceStatus);
			break;
	}
}


//****************************************************************************
// TxD last Status: Sends the checksum
//****************************************************************************
void TxDSendCommandChecksum()
{
	TxDJumpAddress = TxDSendDummy;
	TxDSendByte(RxDChecksum);
}

//****************************************************************************
// TxD dummy status: Next interrupt after checksum-> Lazy down, do nothing  
//****************************************************************************
void TxDSendDummy()
{}




//****************************************************************************
// SetYToPeripherieID: Used by WriteDevice/ ReadDevice
// d0: PeripherieID to find in PeripherieTable
// found-> Y is set to appropriate PeripherieTable entry
// not found-> d0 is 0 after call
// 
//****************************************************************************
PeripherieTableEntry* GetPeripherieTableEntry(uint8_t ID)
{
	PeripherieTableEntry* rval = &PeripherieTable;
	for (int i=0; i< 2; i++)
	{
		if (rval->ID == ID)
			return rval;
		rval++;
	}
	return rval;
}



//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//XXXX    XXXXX     XXX     XXXX     X     X   XXXXX   X     X
//X   X   X        X   X    X   X    XX   XX   X       XX   XX
//X   X   X        X   X    X    X   X X X X   X       X X X X
//XXX     XXXXX    XXXXX    X    X   X  X  X   XXXXX   X  X  X
//X  X    X       X     X   X    X   X     X   X       X     X
//X   X   X       X     X   X   X    X     X   X       X     X
//X   X   XXXXX   X     X   XXXX     X     X   XXXXX   X     X
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
void RxDCommandReadMemGetFirstTargetByte(uint8_t receivedByte)
{
	RxDMemTargetAddress = receivedByte;
	RxDJumpAddress = RxDCommandReadMemGetSecondTargetByte;
}


void RxDCommandReadMemGetSecondTargetByte(uint8_t receivedByte)
{
	RxDMemTargetAddress+= (int) receivedByte * 256;
	RxDJumpAddress = RxDCommandReadMemGetFirstLengthByte;
}



void RxDCommandReadMemGetFirstLengthByte(uint8_t receivedByte)
{
	RxDDataLength = receivedByte;
	RxDJumpAddress = RxDCommandReadMemGetSecondLengthByte;
}


void RxDCommandReadMemGetSecondLengthByte(uint8_t receivedByte)
{
	RxDDataLength+= (int) receivedByte * 256;
	RxDJumpAddress = RxDGetTelegrammID;
	// Prepare and start Answer Telegramm
	TxDJumpAddress = TxDSendCommandAck;
	TxDSendByte(RxDActualTelegrammID);

}




//*******SUB ROUTINES ANSWER TELEGRAM************************************
void TxDSendCommandReadMemData()
{
	if (0!= RxDDataLength)
	{
		RxDDataLength--;
		TxDSendByte(*RxDMemTargetAddress);
		RxDMemTargetAddress++;
	}
	else
		TxDSendCommandChecksum();
}


//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//X         X   XXXX    X   XXXXXXX   XXXXX   X     X   XXXXX   X     X
//X         X   X   X   X      X      X       XX   XX   X       XX   XX
//X    X    X   X   X   X      X      X       X X X X   X       X X X X
//X    X    X   XXXX    X      X      XXXX    X  X  X   XXXXX   X  X  X
// X  X X  X    X  X    X      X      X       X     X   X       X     X
// X X   X X    X   X   X      X      X       X     X   X       X     X
//  X     X     X   X   X      X      XXXXX   X     X   XXXXX   X     X
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
void RxDCommandWriteMemGetFirstTargetByte(uint8_t receivedByte)
{
	RxDMemTargetAddress = receivedByte;
	RxDJumpAddress =  RxDCommandWriteMemGetSecondTargetByte;
}


void RxDCommandWriteMemGetSecondTargetByte(uint8_t receivedByte)
{
	RxDMemTargetAddress+= (int)receivedByte * 256;
	RxDJumpAddress = RxDCommandWriteMemGetFirstLengthByte;
}


void RxDCommandWriteMemGetFirstLengthByte(uint8_t receivedByte)
{	
	RxDDataLength = receivedByte;
	RxDJumpAddress = RxDCommandWriteMemGetSecondLengthByte;
}



void RxDCommandWriteMemGetSecondLengthByte(uint8_t receivedByte)
{
	RxDDataLength += (int)receivedByte * 256;
	RxDJumpAddress = RxDCommandWriteMemGetData;
}


void RxDCommandWriteMemGetData(uint8_t receivedByte)
{
	*RxDMemTargetAddress = receivedByte;
	RxDMemTargetAddress++;

	if (0 != RxDDataLength)
		RxDDataLength--;
	else
	{
		RxDJumpAddress = RxDGetTelegrammID;
		// Prepare and start Answer Telegramm
		TxDJumpAddress=  TxDSendCommandAck;
		TxDSendByte(RxDActualTelegrammID);
	}		
}





//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//XXXX    XXXXX     XXX     XXXX     XXXX      XXXXX   X     X   X    XXXX   XXXXX
//X   X   X        X   X    X   X    X   X     X       X     X   X   X       X
//X   X   X        X   X    X    X   X    X    X       X     X   X   X       X
//XXX     XXXXX    XXXXX    X    X   X    X    XXXXX    X   X    X   X       XXXXX
//X  X    X       X     X   X    X   X    X    X        X   X    X   X       X
//X   X   X       X     X   X   X    X   X     X         X X     X   X       X
//X   X   XXXXX   X     X   XXXX     XXXX      XXXXX      X      X    XXXX   XXXXX
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************


//***********************************************************
//Byte3: PeripherieID expected
//***********************************************************
void RxDCommandReadDeviceGetPeripherieID(uint8_t receivedByte)
{
	RxDPeripherieID = receivedByte;
	RxDJumpAddress = RxDCommandReadDeviceGetFirstLengthByte;
}





void RxDCommandReadDeviceGetFirstLengthByte(uint8_t receivedByte)
{	
	RxDDataLength = receivedByte;
	RxDJumpAddress = RxDCommandReadDeviceGetSecondLengthByte;
}


void RxDCommandReadDeviceGetSecondLengthByte(uint8_t receivedByte)
{
	RxDDataLength+= (int)receivedByte * 256;
	RxDJumpAddress = RxDGetTelegrammID; //No further data in RxD expected



	PeripherieTableEntry* tableEntry = GetPeripherieTableEntry(RxDPeripherieID);
	if (tableEntry < PeripherieTableEndAddress)
	{
		tableEntry->LockByte = 100;
		sei();
		while((tableEntry->LockByte & PT_LOCK_TWICurrentRead_Mask) != 0);



		cli();
		tableEntry->LockByte |= PT_LOCK_RS232CurrentRead_Mask; // Lock it
		RxDMemTargetAddress = tableEntry->Comm_ReadBufferAddress;
		RxDReadWriteDeviceStatus=0;
		// Start answer telegram
		TxDJumpAddress = TxDSendCommandAck;
		TxDSendByte(RxDActualTelegrammID);
	}
}






//*******SUB ROUTINES ANSWER TELEGRAM************************************
void TxDSendCommandReadDeviceData()
{
	if (0 == RxDDataLength)
	{
		PeripherieTableEntry* tableEntry = GetPeripherieTableEntry(RxDPeripherieID);
		if (tableEntry < PeripherieTableEndAddress)
		{
			tableEntry->LockByte &= ~PT_LOCK_RS232CurrentRead_Mask;
		}
		TxDSendCommandChecksum();
	}
	else
	{
		RxDDataLength--;
		TxDSendByte(*RxDMemTargetAddress);
		RxDMemTargetAddress++;
	}
}






//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//X         X   XXXX    X   XXXXXXX   XXXXX   XXXX      XXXXX   X     X   X    XXXX   XXXXX
//X         X   X   X   X      X      X       X   X     X       X     X   X   X       X
//X    X    X   X   X   X      X      X       X    X    X       X     X   X   X       X
//X    X    X   XXXX    X      X      XXXX    X    X    XXXXX    X   X    X   X       XXXXX
// X  X X  X    X  X    X      X      X       X    X    X        X   X    X   X       X
// X X   X X    X   X   X      X      X       X   X     X         X X     X   X       X
//  X     X     X   X   X      X      XXXXX   XXXX      XXXXX      X      X    XXXX   XXXXX
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//***********************************************************
//Byte3: PeripherieID expected
//***********************************************************
void RxDCommandWriteDeviceGetPeripherieID(uint8_t receivedByte)
{
	RxDPeripherieID = receivedByte;
	RxDJumpAddress = RxDCommandWriteDeviceGetFirstLengthByte;
}


//***********************************************************
//Byte4: lower byte length expected
//***********************************************************
void RxDCommandWriteDeviceGetFirstLengthByte(uint8_t receivedByte)
{	
	RxDDataLength = receivedByte;
	RxDJumpAddress = RxDCommandWriteDeviceGetSecondLengthByte;
}


//***********************************************************
//Byte5: upper byte length expected
//***********************************************************
void RxDCommandWriteDeviceGetSecondLengthByte(uint8_t receivedByte)
{
	RxDDataLength+= (int)receivedByte * 256;
	RxDJumpAddress = RxDCommandWriteDeviceGetData;



	PeripherieTableEntry* tableEntry = GetPeripherieTableEntry(RxDPeripherieID);

	if (tableEntry < PeripherieTableEndAddress)
	{
		if ((tableEntry->LockByte & PT_LOCK_TWICurrentWrite_Mask) != 0)
		{
			//Table entry is locked by TWI->Error set in answer telegram
			RxDReadWriteDeviceStatus = MSK_RxD_RWDSError;//Signal Error for answer telegram
		}
		else
		{
			tableEntry->LockByte |= PT_LOCK_RS232CurrentWrite_Mask;
			RxDMemTargetAddress = tableEntry->Comm_WriteBufferAddress;
			RxDReadWriteDeviceStatus=0;//Signal Success for answer telegram
		}

	}
	else
	{
		 // PeripherieID not found->Error set in answer telegram
		RxDReadWriteDeviceStatus = MSK_RxD_RWDSError;//Signal Error for answer telegram
	}


}

//***********************************************************
//Byte6..x: data expected
//***********************************************************
void RxDCommandWriteDeviceGetData(uint8_t receivedByte)
{
	if (RxDReadWriteDeviceStatus == 0)
	{
		// Access granted
		*RxDMemTargetAddress = receivedByte;
		RxDMemTargetAddress++;
	}

	RxDDataLength--;
	if (0== RxDDataLength)
	{
		// last byte
		RxDJumpAddress = RxDGetTelegrammID;
		//Unlock table entry and set NewData Flag
		PeripherieTableEntry* tableEntry = GetPeripherieTableEntry(RxDPeripherieID);
		tableEntry->LockByte &=  ~PT_LOCK_RS232CurrentWrite_Mask;

		// Prepare and start Answer Telegramm
		TxDJumpAddress = TxDSendCommandAck;
		TxDSendByte(RxDActualTelegrammID);
	}
}



void UART_Init()
{
	///***********
	// 9600 Baud
	///***********
	UBRR0H=0;
	UBRR0L=51;

	// Set frame format: 8data, 1 Stop bit, no parity
	UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);

	//Transmitter/Receiver enabled/interrupt enabled
	UCSR0B = (1<<TXCIE0) | (1<<RXCIE0) | (1<<TXEN0) | (1<<RXEN0);	
}




//******************************************
//RxD
//******************************************
ISR(USART_RX_vect)
{
	uint8_t receivedByte = UDR0; // Get received byte in d0
	RxDInterruptService(receivedByte);
}


//*********************************************************************************
//TXD
//*********************************************************************************

void TxDSendByte(uint8_t sendByte)
{
	while ((UCSR0A & (1<<UDRE0)) == 0); // wait for UART data register empty flag
	UDR0 = sendByte;
}


ISR(USART_TX_vect)
{
	TxDInterruptService();
} 




//***************************************************************************
// COMMON
//***************************************************************************

uint8_t MakeChecksumOfByte(uint8_t bt)
{
	uint8_t rval=0;
	for (int i = 0; i < 8; i++)
	{
		rval += (bt &0x01);	
		bt = (bt >> 1);
	}
	return rval;
}







