//***********************************************************************	
// INIT
//***********************************************************************
InitRS232Telegram:
	PrepareJump RxDJumpAddress, RxDGetTelegrammID
	PrepareJump TxDJumpAddress, TxDSendDummy
	ret


//***********************************************************************	
// Called from RS232.asm (RxD Interrupt vector)
//***********************************************************************	
RxDInterruptService:
		call RefreshRS232TimeOutWatchdog
		lds d0, UDR0 // Get received byte in d0
		Jump RxDJumpAddress


//***********************************************************************	
// Called from RS232.asm (TxD Interrupt vector)
//***********************************************************************	
TxDInterruptService:
		call RefreshRS232TimeOutWatchdog
		Jump TxDJumpAddress
		ret // Never called but better here than anywhere



//***********************************************************************	
// RefreshRS232TimeOutWatchdog: 
// Called from subroutines in this module.
// Signals a running data transfer by initializing a watchdog counter
// if watchdog counter (decremented by MainTimer.asm) is timed out, 
//	RxD/TxD state machines are switcheing to 1st State
//***********************************************************************	
RefreshRS232TimeOutWatchdog: 
	push d0
	SetRAMPointer X, RS232TimeOutWatchdog
	ldi d0, 15
	st X, d0
	pop d0
	ret


//***********************************************************************	
// TestRS232TimeOutWatchdog: 
// Called from MainTimerInterrupt: If Counter is timed out, 
//	RxD/TxD state machines are switcheing to 1st State 
//***********************************************************************	
TestRS232TimeOutWatchdog:
	SetRAMPointer X, RS232TimeOutWatchdog
	ld d0, X
	cpi d0, 0 
	breq TestRS232TimeOutWatchdog_End
	dec d0
	brne TestRS232TimeOutWatchdog_End
	call InitRS232Telegram
TestRS232TimeOutWatchdog_End:
	st X, d0
	ret


//********************************************************************
// RxD 1st Status: a TelegramID(TicketID to return in answer) is expected
//********************************************************************
RxDGetTelegrammID:
		sts RxDActualTelegrammID, d0
		PrepareJump RxDJumpAddress, RxDGetCommand
		ret

//********************************************************************
// RxD 2nd Status: a Command is expected:
//	1 = WriteMem
//	2 = ReadMem
//  3 = WriteDevice	
//	4 = ReadDevice
//********************************************************************
RxDGetCommand:
		sts RxDActualCommand, d0 // Save actual command
		//Switch command table
		cpi d0, 1 // WriteMem
		breq RxDGetCommandWriteMem
		cpi d0, 2 // ReadMem
		breq RxDGetCommandReadMem
		cpi d0, 3 // WriteDevice	
		breq RxDGetCommandWriteDevice
		cpi d0, 4 //ReadDevice
		breq RxDGetCommandReadDevice
		// not ALLOWED Commands
		PrepareJump RxDJumpAddress, RxDGetTelegrammID
		// Todo Sending Error Telegram		
		ret
RxDGetCommandWriteMem:
	PrepareJump RxDJumpAddress, RxDCommandWriteMemGetFirstTargetByte			
	ret		
RxDGetCommandReadMem:
	PrepareJump RxDJumpAddress, RxDCommandReadMemGetFirstTargetByte
	ret
RxDGetCommandWriteDevice:
	PrepareJump RxDJumpAddress, RxDCommandWriteDeviceGetPeripherieID
	ret
RxDGetCommandReadDevice:
	PrepareJump RxDJumpAddress, RxDCommandReadDeviceGetPeripherieID
	ret


RxDGetCommand_End:



//********************************************************************
// TxD 2nd Status: decision what kind of answer is returned.
// Remarks: 1st Status (answering with telegramID) is done prior 
//	by RxD subroutine)
//********************************************************************
TxDSendCommandAck:
	lds d0, RxDActualCommand
	cpi d0, 0x01 //WriteMem
	breq TxDSendCommandAckWriteMem	

	cpi d0, 0x02 //ReadMem
	breq TxDSendCommandAckReadMem	

	cpi d0, 0x03 //WriteDevice
	breq TxDSendCommandAckWriteDevice

	cpi d0, 0x04 //ReadDevice
	breq TxDSendCommandAckReadDevice
	ret

TxDSendCommandAckWriteMem:
	PrepareJump TxDJumpAddress, TxDSendCommandChecksum
	ldi d0, 0x81
	call TxDSendByte
	ret

TxDSendCommandAckReadMem:	
	PrepareJump TxDJumpAddress, TxDSendCommandReadMemData
	ldi d0, 0x82
	call TxDSendByte
	ret

TxDSendCommandAckWriteDevice:
	PrepareJump TxDJumpAddress, TxDSendCommandChecksum
	ldi d0, 0x83
	lds d1, RxDReadWriteDeviceStatus // Set Status generated in RxDCommandReadDevice...
	or d0, d1
	call TxDSendByte
	ret


TxDSendCommandAckReadDevice:
	PrepareJump TxDJumpAddress, TxDSendCommandReadDeviceData
	ldi d0, 0x84
	lds d1, RxDReadWriteDeviceStatus // Set Status generated in RxDCommandReadDevice...
	or d0, d1
	call TxDSendByte
	ret



//****************************************************************************
// TxD last Status: Sends the checksum
//****************************************************************************
TxDSendCommandChecksum:
	PrepareJump TxDJumpAddress, TxDSendDummy
	lds d0, RxDChecksum
	call TxDSendByte
	ret	


//****************************************************************************
// TxD dummy status: Next interrupt after checksum-> Lazy down, do nothing  
//****************************************************************************
TxDSendDummy:
	//nothing to do
	ret



//****************************************************************************
// SetYToPeripherieID: Used by WriteDevice/ ReadDevice
// d0: PeripherieID to find in PeripherieTable
// found-> Y is set to appropriate PeripherieTable entry
// not found-> d0 is 0 after call
// 
//****************************************************************************
SetYToPeripherieID:
	ldi YH, high(PeripherieTable)
	ldi YL, low(PeripherieTable)

SetYToPeripherieID_FindTableEntry:

		
RxDCommandReadDevice_FindTableEntry_Loop:
	ldd d1, Y+PT_OFFS_PeriId // Compare PeripherieID
	cp d0, d1
	breq SetYToPeripherieID_TableEntryFound
	adiw Y, PT_OFFS_TableEntry
	cpi YL, low(PeripherieTable_End)
	brlo RxDCommandReadDevice_FindTableEntry_Loop
RxDCommandReadDevice_FindTableEntry_Loop_End:
	clr d0
	ret  // If Peripherie not found return with d0 = 0

SetYToPeripherieID_TableEntryFound:
	ret




//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//XXXX    XXXXX     XXX     XXXX     X     X   XXXXX   X     X
//X   X   X        X   X    X   X    XX   XX   X       XX   XX
//X   X   X        X   X    X    X   X X X X   X       X X X X
//XXX     XXXXX    XXXXX    X    X   X  X  X   XXXXX   X  X  X
//X  X    X       X     X   X    X   X     X   X       X     X
//X   X   X       X     X   X   X    X     X   X       X     X
//X   X   XXXXX   X     X   XXXX     X     X   XXXXX   X     X
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
RxDCommandReadMemGetFirstTargetByte:
	sts RxDMemTargetAddress, d0
	PrepareJump RxDJumpAddress, RxDCommandReadMemGetSecondTargetByte
	ret

RxDCommandReadMemGetSecondTargetByte:
	sts RxDMemTargetAddress+1, d0
	PrepareJump RxDJumpAddress, RxDCommandReadMemGetFirstLengthByte
	ret

RxDCommandReadMemGetFirstLengthByte:
	sts RxDDataLength, d0
	PrepareJump RxDJumpAddress, RxDCommandReadMemGetSecondLengthByte
	ret

RxDCommandReadMemGetSecondLengthByte:
	sts RxDDataLength+1, d0
	PrepareJump RxDJumpAddress, RxDGetTelegrammID //No further data in RxD expected
	
	// Prepare and start Answer Telegramm
	PrepareJump TxDJumpAddress, TxDSendCommandAck
	lds d0, RxDActualTelegrammID
	call TxDSendByte

	ret

//*******SUB ROUTINES ANSWER TELEGRAM************************************
TxDSendCommandReadMemData:
	Mac_Decrement16BitRamCounterIfNotNull RxDDataLength
	cpi d0, 0
	breq TxDSendCommandReadMemData_SendChecksum//TxDSendCommandChecksum

	lds XL, RxDMemTargetAddress
	lds XH, RxDMemTargetAddress+1
	ld d0, X+
	sts RxDMemTargetAddress, XL
	sts RxDMemTargetAddress +1 , XH
	call TxDSendByte
	ret

TxDSendCommandReadMemData_SendChecksum:
	jmp TxDSendCommandChecksum
	ret

TxDSendCommandReadMemData_End:



//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//X         X   XXXX    X   XXXXXXX   XXXXX   X     X   XXXXX   X     X
//X         X   X   X   X      X      X       XX   XX   X       XX   XX
//X    X    X   X   X   X      X      X       X X X X   X       X X X X
//X    X    X   XXXX    X      X      XXXX    X  X  X   XXXXX   X  X  X
// X  X X  X    X  X    X      X      X       X     X   X       X     X
// X X   X X    X   X   X      X      X       X     X   X       X     X
//  X     X     X   X   X      X      XXXXX   X     X   XXXXX   X     X
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
//**************************************************************************************
RxDCommandWriteMemGetFirstTargetByte:
	sts RxDMemTargetAddress, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteMemGetSecondTargetByte
	ret

RxDCommandWriteMemGetSecondTargetByte:
	sts RxDMemTargetAddress+1, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteMemGetFirstLengthByte
	ret

RxDCommandWriteMemGetFirstLengthByte:
	sts RxDDataLength, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteMemGetSecondLengthByte
	ret

RxDCommandWriteMemGetSecondLengthByte:
	sts RxDDataLength+1, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteMemGetData
	ret

RxDCommandWriteMemGetData:
	lds XL, RxDMemTargetAddress
	lds XH, RxDMemTargetAddress+1
	st X+, d0
	sts RxDMemTargetAddress, XL
	sts RxDMemTargetAddress +1 , XH



	Mac_Decrement16BitRamCounterIfNotNull RxDDataLength
	cpi d0, 0
	breq RxDCommandWriteMemGetData_LastByte
	ret

RxDCommandWriteMemGetData_LastByte:
	PrepareJump RxDJumpAddress, RxDGetTelegrammID
	
	// Prepare and start Answer Telegramm
	PrepareJump TxDJumpAddress, TxDSendCommandAck
	lds d0, RxDActualTelegrammID
	call TxDSendByte
	ret






//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//XXXX    XXXXX     XXX     XXXX     XXXX      XXXXX   X     X   X    XXXX   XXXXX
//X   X   X        X   X    X   X    X   X     X       X     X   X   X       X
//X   X   X        X   X    X    X   X    X    X       X     X   X   X       X
//XXX     XXXXX    XXXXX    X    X   X    X    XXXXX    X   X    X   X       XXXXX
//X  X    X       X     X   X    X   X    X    X        X   X    X   X       X
//X   X   X       X     X   X   X    X   X     X         X X     X   X       X
//X   X   XXXXX   X     X   XXXX     XXXX      XXXXX      X      X    XXXX   XXXXX
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************


//***********************************************************
//Byte3: PeripherieID expected
//***********************************************************
RxDCommandReadDeviceGetPeripherieID:
	sts RxDPeripherieID, d0
	PrepareJump RxDJumpAddress, RxDCommandReadDeviceGetFirstLengthByte
	ret



//***********************************************************
//Byte4: Upper byte length expected
//***********************************************************
RxDCommandReadDeviceGetFirstLengthByte:
	sts RxDDataLength, d0
	PrepareJump RxDJumpAddress, RxDCommandReadDeviceGetSecondLengthByte
	ret



//***********************************************************
//Byte5: lower byte length expected
//***********************************************************
RxDCommandReadDeviceGetSecondLengthByte:
	sts RxDDataLength+1, d0
	PrepareJump RxDJumpAddress, RxDGetTelegrammID //No further data in RxD expected

	lds d0, RxDPeripherieID
	call SetYToPeripherieID

	cpi d0, 0
	brne RxDCommandReadDevice_TableEntryFound

	ret // PeripherieID not found-> no answer telegram is generated

RxDCommandReadDevice_TableEntryFound:
RxDCommandReadDevice_TableEntryFound_WaitForUnlockLoop:
	sei //Enable further interrupts

	ldd d0, Y+PT_OFFS_LOCK
	andi d0, PT_LOCK_TWICurrentRead_Mask
	brne RxDCommandReadDevice_TableEntryFound_WaitForUnlockLoop
RxDCommandReadDevice_TableEntryFound_WaitForUnlockLoop_End:
	cli 

	ldd d0, Y+PT_OFFS_LOCK
	ori d0, PT_LOCK_RS232CurrentRead_Mask
	std Y+PT_OFFS_LOCK, d0

	//Copy Address of communicators Pewripherie Readmem to RxDMemTargetAddress
	ldd d0, Y+PT_OFFS_ReadMem_Addr
	sts RxDMemTargetAddress, d0
	ldd d0, Y+(PT_OFFS_ReadMem_Addr+1)
	sts RxDMemTargetAddress +1, d0
	clr d0
	sts RxDReadWriteDeviceStatus, d0 //Signal Success for answer telegram
	

	// Start answer telegram
	PrepareJump TxDJumpAddress, TxDSendCommandAck
	lds d0, RxDActualTelegrammID
	call TxDSendByte

	ret

//*******SUB ROUTINES ANSWER TELEGRAM************************************
TxDSendCommandReadDeviceData:
	Mac_Decrement16BitRamCounterIfNotNull RxDDataLength
	cpi d0, 0
	breq TxDSendCommandReadDeviceData_SendChecksum//TxDSendCommandChecksum

	lds XL, RxDMemTargetAddress
	lds XH, RxDMemTargetAddress+1
	ld d0, X+
	sts RxDMemTargetAddress, XL
	sts RxDMemTargetAddress +1 , XH
	call TxDSendByte
	ret

TxDSendCommandReadDeviceData_SendChecksum:
	lds d0, RxDPeripherieID
	call SetYToPeripherieID
	cpi d0, 0
	breq TxDSendCommandReadDeviceData_Unlock_End // Nothing to unlock?!?

TxDSendCommandReadDeviceData_Unlock:
	ldd d0, Y+PT_OFFS_LOCK
	andi d0, ~(PT_LOCK_Rs232CurrentRead_Mask)
	std Y+PT_OFFS_LOCK, d0
TxDSendCommandReadDeviceData_Unlock_End:
	jmp TxDSendCommandChecksum

	ret
TxDSendCommandReadDeviceData_End:







//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//X         X   XXXX    X   XXXXXXX   XXXXX   XXXX      XXXXX   X     X   X    XXXX   XXXXX
//X         X   X   X   X      X      X       X   X     X       X     X   X   X       X
//X    X    X   X   X   X      X      X       X    X    X       X     X   X   X       X
//X    X    X   XXXX    X      X      XXXX    X    X    XXXXX    X   X    X   X       XXXXX
// X  X X  X    X  X    X      X      X       X    X    X        X   X    X   X       X
// X X   X X    X   X   X      X      X       X   X     X         X X     X   X       X
//  X     X     X   X   X      X      XXXXX   XXXX      XXXXX      X      X    XXXX   XXXXX
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
//***********************************************************
//Byte3: PeripherieID expected
//***********************************************************
RxDCommandWriteDeviceGetPeripherieID:
	sts RxDPeripherieID, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteDeviceGetFirstLengthByte
	ret



//***********************************************************
//Byte4: lower byte length expected
//***********************************************************
RxDCommandWriteDeviceGetFirstLengthByte:
	sts RxDDataLength, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteDeviceGetSecondLengthByte
	ret



//***********************************************************
//Byte5: upper byte length expected
//***********************************************************
RxDCommandWriteDeviceGetSecondLengthByte:
	sts RxDDataLength+1, d0
	PrepareJump RxDJumpAddress, RxDCommandWriteDeviceGetData 

	lds d0, RxDPeripherieID
	call SetYToPeripherieID

	cpi d0, 0
	brne RxDCommandWriteDevice_TableEntryFound

	ldi d0, MSK_RxD_RWDSError // PeripherieID not found->Error set in answer telegram
	sts RxDReadWriteDeviceStatus, d0 //Signal Error for answer telegram
	ret 

RxDCommandWriteDevice_TableEntryFound:
//Try to lock table entry
	ldd d0, Y+PT_OFFS_LOCK
	mov d1, d0
	andi d1, PT_LOCK_TWICurrentWrite_Mask
	breq RxDCommandWriteDevice_UnlockedTableEntryFound_LockIt

		
	ldi d0, MSK_RxD_RWDSError //Table entry is locked by TWI->Error set in answer telegram
	sts RxDReadWriteDeviceStatus, d0 //Signal Error for answer telegram
	ret

RxDCommandWriteDevice_UnlockedTableEntryFound_LockIt:
	ori d0, PT_LOCK_RS232CurrentWrite_Mask
	std Y+PT_OFFS_LOCK, d0

	//Copy Address of communicators Peripherie WriteMem to RxDMemTargetAddress
	ldd d0, Y+PT_OFFS_WriteMem_Addr
	sts RxDMemTargetAddress, d0
	ldd d0, Y+(PT_OFFS_WriteMem_Addr+1)
	sts RxDMemTargetAddress +1, d0
	clr d0
	sts RxDReadWriteDeviceStatus, d0 //Signal Success for answer telegram
	ret


//***********************************************************
//Byte6..x: data expected
//***********************************************************
RxDCommandWriteDeviceGetData:
	mov d1, d0
	lds d0, RxDReadWriteDeviceStatus //PeripherieID not found or Table entry locked?-> Do NOT Write data
	cpi d0, 0
	brne RxDCommandWriteDeviceGetData_Write_End 

RxDCommandWriteDeviceGetData_Write:
	lds XL, RxDMemTargetAddress
	lds XH, RxDMemTargetAddress+1
	st X+, d1
	sts RxDMemTargetAddress, XL
	sts RxDMemTargetAddress +1 , XH

RxDCommandWriteDeviceGetData_Write_End:
	Mac_Decrement16BitRamCounterUntilNull RxDDataLength
	//Mac_Decrement16BitRamCounterIfNotNull RxDDataLength
	cpi d0, 0
	breq RxDCommandWriteDeviceGetData_LastByte
	ret

RxDCommandWriteDeviceGetData_LastByte:
	PrepareJump RxDJumpAddress, RxDGetTelegrammID

	//Unlock table entry and set NewData Flag
	lds d0, RxDPeripherieID
	call SetYToPeripherieID
	ldd d0, Y+PT_OFFS_LOCK
	cbr d0, PT_LOCK_RS232CurrentWrite_Mask 
	sbr d0, PT_LOCK_NewWriteData_Mask  
	std Y+PT_OFFS_LOCK, d0


	
	// Prepare and start Answer Telegramm
	PrepareJump TxDJumpAddress, TxDSendCommandAck
	lds d0, RxDActualTelegrammID
	call TxDSendByte
	ret



